/*
 
Charles K. Fisher, 6/25/2014
 
REQUIRES GNU SCIENTIFIC LIBRARIES (GSL) and OpenMP
 
*/

// Header files

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <time.h>
#include <vector>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_statistics.h>
#include <omp.h>

// Compile command on my computer (path to /include and /lib will be different on different computers)
// gcc name.cpp -fopenmp -o name -I/opt/local/include -L/opt/local/lib -lgsl -lgslcblas -lm

using namespace std;

// Class definitions

// Main program
int main()
{
    omp_set_num_threads(6);
    
    int n = 200, p = 28395; // number of samples, and number of features
    
    // read in data from files
    
    gsl_vector * y = gsl_vector_alloc(n);
    gsl_matrix * X = gsl_matrix_alloc(n,p);
    
    FILE * file1 = fopen("soybean_response.dat","r");
    gsl_vector_fscanf(file1,y);
    fclose(file1);
    
    FILE * file2 = fopen("soybean_features.dat","r");
    gsl_matrix_fscanf(file2,X);
    fclose(file2);
    
    // compute rYX
    gsl_vector * rYX = gsl_vector_alloc(p);
    for(int j = 0; j < p; j++)
    {
        double tmp = 0.0;
        for(int l = 0; l < n; l++)
        {
            tmp += gsl_vector_get(y,l)*gsl_matrix_get(X,l,j);
        }
        gsl_vector_set(rYX,j,tmp/(n-1.0));
    }
    
    // compute rXX
    double r = 0.0;
    gsl_matrix * J = gsl_matrix_alloc(p,p);
    #pragma omp parallel for
    for(int i = 0; i < p; i++)
    {
        for(int j = i; j < p; j++)
        {
            double rXX = 0.0;
            for(int l = 0; l < n; l++)
            {
                rXX += gsl_matrix_get(X,l,i)*gsl_matrix_get(X,l,j);
            }
            rXX = rXX / (n-1.0);
            if(i != j)
            {
                #pragma omp atomic
                r += pow(rXX,2.0);
            }
            
            gsl_matrix_set(J,i,j,pow(rXX,2.0) / n - rXX*gsl_vector_get(rYX,i)*gsl_vector_get(rYX,j) + 0.5*pow(gsl_vector_get(rYX,i),2.0)*pow(gsl_vector_get(rYX,j),2.0));
            gsl_matrix_set(J,j,i,gsl_matrix_get(J,i,j));
        }
    }
    gsl_vector_free(y);
    gsl_matrix_free(X);
    
    // compute critical penality parameter (called "scale" here)
    double scale = n*(1.0+p*sqrt(2.0 * r  / ( p*(p-1.0))));
    cout << sqrt(2.0*r / ( p*(p-1.0))) << " " << 1.0/scale << " " << 1.0/(n*p) << endl;
    
    // output file
    ofstream output;
    output.open("soyean_selection_path.dat");
    
    gsl_vector * m1 = gsl_vector_alloc(p);
    gsl_vector * m2 = gsl_vector_alloc(p);
    gsl_vector_set_zero(m1);
    gsl_vector_set_zero(m2);
    
    // compute variable selection path
    double delta = 0.05 / (scale), epsilon = 0.0;
    for(int t = 0; t<60; t++)
    {
        epsilon += delta;
        #pragma omp parallel for
        for(int i = 0; i < p; i++)
        {
            double tmp = 0.0;
            for(int j = 0; j < p; j++)
            {
                if(j!=i)
                {
                    tmp += gsl_matrix_get(J,i,j)*(1.0+0.5*gsl_vector_get(m1,j));
                }
            }
            gsl_vector_set(m2,i,tanh(0.25*epsilon*pow(n,2.0)*(pow(gsl_vector_get(rYX,i),2.0) - 1.0/n + n*epsilon*(1.0/n - pow(gsl_vector_get(rYX,i),2.0) + 0.5*pow(gsl_vector_get(rYX,i),4.0)) + n*epsilon*tmp)));
        }
        
        gsl_vector_memcpy(m1,m2);
    
        // output to file
        output << epsilon << " ";
        for(int i = 0; i < p; i++)
        {
            output << setprecision(6) << gsl_vector_get(m2,i) << " ";
        }
        output << endl;
    }
    
    output.close();
    
    // free memory
    gsl_vector_free(rYX);
    gsl_matrix_free(J);
    gsl_vector_free(m1);
    gsl_vector_free(m2);

    return 0;
}
